@@ -3,20 +3,24 @@ from __future__ import annotations
 import asyncio
 import binascii
 import importlib
+import io
 import os
+import pty
 import ssl
 import subprocess
 import sys
+import tarfile
 import tempfile
-from dataclasses import dataclass, field
 from pathlib import Path
 
 import aiofiles
+import aioshutil
+from aiofiles import os as aiofiles_os
 from aiohttp import ClientSession
-from aiotinyrpc.auth import SignatureAuthProvider
-from aiotinyrpc.protocols.jsonrpc import JSONRPCProtocol
-from aiotinyrpc.server import RPCServer
-from aiotinyrpc.transports.socket import EncryptedSocketServerTransport
+from fluxrpc.auth import SignatureAuthProvider
+from fluxrpc.protocols.jsonrpc import JSONRPCProtocol
+from fluxrpc.server import RPCServer
+from fluxrpc.transports.socket.server import EncryptedSocketServerTransport
 from cryptography import x509
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.asymmetric import rsa
@@ -71,6 +75,8 @@ class FluxAgent:
         self.primary_agent = primary_agent
         self.component_name, self.app_name = get_app_and_component_name()
 
+        log.info(f"Component name: {self.component_name}, App name: {self.app_name}")
+
         if not self.signed_vault_connections and not self.verify_source_address:
             # Must verify source address as a minimum
             self.verify_source_address = True
@@ -97,7 +103,6 @@ class FluxAgent:
             async with session.get(
                 f"https://api.runonflux.io/apps/appowner?appname={app_name}"
             ) as resp:
-                # print(resp.status)
                 data = await resp.json()
                 zelid = data.get("data", "")
         return zelid
@@ -132,8 +137,9 @@ class FluxAgent:
             raise ValueError("Primary agent must be provided if subordinate")
 
     def register_extensions(self):
-        self.extensions.add_method(self.get_all_files_crc)
-        self.extensions.add_method(self.write_files)
+        self.extensions.add_method(self.get_all_object_hashes)
+        self.extensions.add_method(self.write_objects)
+        self.extensions.add_method(self.remove_objects)
         self.extensions.add_method(self.get_methods)
         self.extensions.add_method(self.get_subagents)
         self.extensions.add_method(self.generate_csr)
@@ -142,6 +148,10 @@ class FluxAgent:
         self.extensions.add_method(self.upgrade_to_ssl)
         self.extensions.add_method(self.load_plugins)
         self.extensions.add_method(self.list_server_details)
+        self.extensions.add_method(self.connect_shell)
+        self.extensions.add_method(self.disconnect_shell)
+        self.extensions.add_method(self.get_state)
+        self.extensions.add_method(self.get_directory_hashes)
 
         # self.extensions.add_method(self.run_entrypoint)
         # self.extensions.add_method(self.extract_tar)
@@ -165,11 +175,16 @@ class FluxAgent:
                 f"Sub agent http server running on port {self.registrar.bind_port}"
             )
 
-        self.loop.create_task(self.rpc_server.serve_forever())
+        task = self.loop.create_task(self.rpc_server.serve_forever())
 
         try:
             self.loop.run_forever()
-        finally:
+        except Exception as e:
+            # should this be finally: ?
+            # this is when child fork closes (no loop)
+            # OSError
+            task.cancel()
+        else:
             if self.enable_registrar:
                 self.loop.run_until_complete(self.registrar.cleanup())
 
@@ -199,7 +214,7 @@ class FluxAgent:
         context.load_verify_locations(cafile=ca_cert.name)
         context.check_hostname = False
         context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
-        context.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384")
+        # context.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384")
 
         cert.close()
         key.close()
@@ -213,11 +228,10 @@ class FluxAgent:
             auth_provider=self.auth_provider,
             ssl=context,
         )
-
         await self.rpc_server.transport.stop_server()
         log.info("Non SSL RPC server stopped")
-        self.rpc_server1 = RPCServer(transport, JSONRPCProtocol(), self.extensions)
-        self.loop.create_task(self.rpc_server1.serve_forever())
+        self.rpc_server = RPCServer(transport, JSONRPCProtocol(), self.extensions)
+        self.loop.create_task(self.rpc_server.serve_forever())
 
     def cleanup(self):
         # ToDo: look at cleanup for rpc server too
@@ -234,54 +248,137 @@ class FluxAgent:
         """Returns methods available for the keeper to call"""
         return {k: v.__doc__ for k, v in self.extensions.method_map.items()}
 
-    async def get_file_crc(self, path: str) -> dict:
-        """Open the file and compute the crc, set crc=0 if not found"""
-        # ToDo: catch file PermissionError
+    def get_state(self) -> dict:
+        methods = self.get_methods()
+        plugins = self.extensions.list_plugins()
+        primary_agent = self.primary_agent.to_dict() if self.primary_agent else None
+        return {
+            "methods": methods,
+            "plugins": plugins,
+            "component_name": self.component_name,
+            "app_name": self.app_name,
+            "enable_registrar": self.enable_registrar,
+            "zelid": self.zelid,
+            "working_dir": self.working_dir,
+            "subordinate": self.subordinate,
+            "signed_vault_connections": self.signed_vault_connections,
+            "bind_address": self.bind_address,
+            "bind_port": self.bind_port,
+            "whitelisted_addresses": self.whitelisted_addresses,
+            "verify_source_address": self.verify_source_address,
+            "primary_agent": primary_agent,
+        }
+
+    async def crc_file(self, filename: Path, crc: int) -> int:
+        async with aiofiles.open(filename, "rb") as f:
+            data = await f.read()
+            crc = binascii.crc32(data, crc)
+
+        return crc
+
+    async def crc_directory(self, directory: Path, crc: int) -> int:
+        crc = binascii.crc32(directory.name.encode(), crc)
+        for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
+            crc = binascii.crc32(path.name.encode(), crc)
+
+            if path.is_file():
+                crc = await self.crc_file(path, crc)
+            elif path.is_dir():
+                crc = await self.crc_directory(path, crc)
+        return crc
+
+    async def get_object_crc(self, path: str) -> int:
         p = Path(path)
 
         if not p.is_absolute():
             p = self.working_dir / p
 
-        try:
-            # Todo: brittle as
-            async with aiofiles.open(p, mode="rb") as file:
-                content = await file.read()
-                file.close()
-                crc = binascii.crc32(content)
-        except FileNotFoundError:
-            log.info(f"Local file {p} not found")
-            crc = 0
-        # ToDo: Fix this
-        except Exception as e:
-            log.error(repr(e))
+        if not p.exists():
             crc = 0
 
+        elif p.is_dir():
+            crc = await self.crc_directory(p, 0)
+
+        elif p.is_file():
+            crc = await self.crc_file(p, 0)
+
         return {"name": path, "crc32": crc}
 
-    async def get_all_files_crc(self, files) -> list:
-        """Returns the crc32 for each file that is being managed"""
-        log.info("Returning all vault file hashes")
+    async def get_all_object_hashes(self, objects: list) -> list:
+        """Returns the crc32 for each object that is being managed"""
+        log.info(f"Returning crc's for {len(objects)} object(s)")
         tasks = []
-        for file in files:
-            tasks.append(self.loop.create_task(self.get_file_crc(file)))
+        for obj in objects:
+            tasks.append(self.loop.create_task(self.get_object_crc(obj)))
         results = await asyncio.gather(*tasks)
         return results
 
-    async def write_file(self, path: str, data: str | bytes, executable: bool = False):
-        """Write a single file to disk"""
+    async def get_file_hash(self, file: Path):
+        crc = await self.crc_file(file, 0)
+        return {str(file): crc}
+
+    async def get_directory_hashes(self, dir: str):
+        """Hashes up all files in a specific directory. if
+        give relative path, out working dir is base path. Need
+        to remove this again for each hash to give back common path format"""
+        hashes = {}
+        p = Path(dir)
+        path_relative_to_workdir = False
+        try:
+            if not p.is_absolute():
+                p = self.working_dir / p
+                path_relative_to_workdir = True
+
+            if not p.exists():
+                return hashes
+
+            crc = binascii.crc32(p.name.encode())
+
+            hashes.update({str(p): crc})
+            for path in sorted(p.iterdir(), key=lambda p: str(p).lower()):
+                if path.is_dir():
+                    hashes.update(await self.get_directory_hashes(str(path)))
+
+                elif path.is_file():
+                    hashes.update(await self.get_file_hash(path))
+
+            if path_relative_to_workdir:
+                hashes = {
+                    str(Path(k).relative_to(self.working_dir)): v
+                    for k, v in hashes.items()
+                }
+        except Exception as e:
+            print(repr(e))
+            raise
+
+        return hashes
+
+    async def remove_object(self, obj: str):
+        p = Path(obj)
+
+        if not p.is_absolute():
+            p = self.working_dir / p
+
+        if p.exists():
+            if p.is_dir():
+                await aioshutil.rmtree(p)
+            elif p.is_file():
+                await aiofiles_os.remove(p)
+
+    async def write_object(self, obj: dict):
         # ToDo: brittle file path
         # ToDo: catch file PermissionError etc
 
-        # os.umask(0) - set this is for everyone but doesn't really matter here
+        executable = False  # pass this in dict in future
 
-        if isinstance(data, bytes):
+        if isinstance(obj["data"], bytes):
             mode = "wb"
-        elif isinstance(data, str):
+        elif isinstance(obj["data"], str):
             mode = "w"
         else:
             raise ValueError("Data written must be either str or bytes")
 
-        p = Path(path)
+        p = Path(obj["path"])
 
         if not p.is_absolute():
             p = self.working_dir / p
@@ -290,12 +387,28 @@ class FluxAgent:
 
         # this will make the file being written executable
         opener = self.opener if executable else None
-        try:
-            async with aiofiles.open(p, mode=mode, opener=opener) as file:
-                await file.write(data)
-        # ToDo: tighten this up
-        except Exception as e:
-            log.error(repr(e))
+
+        if obj["is_dir"]:
+            p.mkdir(parents=True, exist_ok=True)
+            if not obj["data"]:
+                return
+
+        if obj.get("uncompressed", False):
+            try:
+                async with aiofiles.open(p, mode=mode, opener=opener) as file:
+                    await file.write(obj["data"])
+            # ToDo: tighten this up
+            except Exception as e:
+                log.error(repr(e))
+
+        else:  # tarball
+            fh = io.BytesIO(obj["data"])
+            try:
+                with tarfile.open(fileobj=fh, mode="r|bz2") as tar:
+                    tar.extractall(str(p))
+                return
+            except Exception as e:
+                print(f"Tarfile error: {repr(e)}")
 
     async def get_subagents(self):
         agents = {}
@@ -303,13 +416,55 @@ class FluxAgent:
             agents = {v.dns_name: v.as_dict() for v in self.registrar.sub_agents}
         return {"sub_agents": agents}
 
-    async def write_files(self, files: dict):
+    async def write_objects(self, objects: list):
         """Will write to disk any file provided, in the format {"name": <content>}"""
         # ToDo: this should be tasks
-        for name, data in files.items():
-            log.info(f"Writing file {name}")
-            await self.write_file(name, data)
-            log.info("Writing complete")
+        for obj in objects:
+            log.info(f"Writing object {obj['path']}")
+            await self.write_object(obj)
+
+    async def remove_objects(self, objects: list):
+        for obj in objects:
+            log.info(f"Removing object {obj}")
+            await self.remove_object(obj)
+
+    async def connect_shell(self, peer):
+        # peer is the source ip, host
+        # json converts tuple to list
+        peer = tuple(peer)
+
+        child_pid, fd = pty.fork()
+        if child_pid == 0:
+
+            # Child process
+            while log.hasHandlers():
+                log.removeHandler(log.handlers[0])
+
+            # trying to suppress log messages?!? Suspect it is this process
+            # somehow sending data down the pipe to the remote SSL end causing
+            # problems decoding ssl every once in a while
+            try:
+                subprocess.run("zsh")
+            except:
+                pass
+
+        else:
+            # Parent process
+
+            # the peer may be the jumphost instead of actual browser,
+            # we have no way of identifying if they are a jumphost so we
+            # pass the RPCClient id of the originating request
+            try:
+                self.rpc_server.transport.attach_pty(child_pid, fd, peer)
+                await self.rpc_server.transport.proxy_pty(peer)
+            except Exception as e:
+                print("In connect_shell")
+                print(repr(e))
+
+    async def disconnect_shell(self, peer):
+        log.info(f"Disconnecting shell for peer: {peer[0]}")
+        peer = tuple(peer)  # json convert to list
+        self.rpc_server.transport.detach_pty(peer)
 
     def list_server_details(self):
         return {
@@ -367,9 +522,6 @@ class FluxAgent:
                 log.error("Plugin load error... skipping")
 
     async def generate_csr(self):
-        if not self.component_name or not self.app_name:
-            self.component_name = "component1"
-            self.app_name = "testapp"
         # ToDo: this needs to include a Fluxnode identifier
         altname = f"{self.component_name}.{self.app_name}.com"
 
@@ -415,6 +567,7 @@ class FluxAgent:
         ).value.get_values_for_type(x509.DNSName)
         log.info(f"Installing cert from issuer {issuer} with Alt names {san}")
 
+        # ToDo: this timing seems a bit off?
         await self.sub_agent.update_local_agent(enrolled=True)
 
     async def upgrade_connection(self):
@@ -425,9 +578,6 @@ class FluxAgent:
         self.ca_cert = cert_bytes
 
     def extract_tar(self, file, target_dir):
-        import tarfile
-        from pathlib import Path
-
         Path(target_dir).mkdir(parents=True, exist_ok=True)
 
         try:
